# SPDX-FileCopyrightText: 2024 Damien Picard
#
# SPDX-License-Identifier: GPL-3.0-or-later


import bpy
import pygments
import pygments.lexers
import pygments.formatter
import os


class BlenderFormatter(pygments.formatter.Formatter):
    def __init__(self, obj, **options):
        pygments.formatter.Formatter.__init__(self, **options)

        self.obj = obj
        self.style_name = options["style"]
        self.styles = {}
        for token, style in self.style:
            self.styles[token] = style

    @staticmethod
    def srgb_to_linearrgb(c):
        """From https://blender.stackexchange.com/a/158902"""
        if c < 0:
            return 0
        if c < 0.04045:
            return c / 12.92
        return ((c + 0.055) / 1.055) ** 2.4

    def hex_to_rgb(self, h, alpha=1):
        """From https://blender.stackexchange.com/a/158902"""
        h = int(h, base=16)
        r = (h & 0xff0000) >> 16
        g = (h & 0x00ff00) >> 8
        b = (h & 0x0000ff)
        return tuple([self.srgb_to_linearrgb(c / 0xff) for c in (r, g, b)] + [alpha])

    def get_material(self, token, token_name, style):
        """Get or create a material for this token and style."""
        mat = bpy.data.materials.get(token_name)
        if mat is None:
            mat = bpy.data.materials.new(token_name)
            mat.use_nodes = True
            mat.node_tree.nodes.remove(mat.node_tree.nodes[0])
            bsdf_node = mat.node_tree.nodes.new("ShaderNodeGroup")
            bsdf_node.location = (10.0, 300.0)
            bsdf_node.node_tree = self.get_shader_node_group()

            mat.node_tree.links.new(
                mat.node_tree.nodes[0].inputs[0], bsdf_node.outputs[0])

            if style["color"] is None:
                color = (0.0, 0.0, 0.0, 0.0)
            else:
                color = self.hex_to_rgb(style["color"])
            bsdf_node.inputs["Color"].default_value = color
            mat.diffuse_color = color
        return mat

    def get_material_slot(self, token):
        """Get or create a material slot for this token."""
        token_name = f"{self.style_name}.{token}"
        slot = self.obj.material_slots.get(token_name)
        if slot is None:
            style = self.styles[token]
            mat = self.get_material(token, token_name, style)
            self.obj.data.materials.append(mat)
            slot = self.obj.material_slots[-1]
        return slot

    def get_shader_node_group(self):
        """Get or create the shader node group abstracting the BSDF."""
        node_group = bpy.data.node_groups.get("Highlighter")
        if node_group is None:
            node_group = bpy.data.node_groups.new("Highlighter", 'ShaderNodeTree')
            node_group.interface.new_socket("BSDF", in_out='OUTPUT', socket_type='NodeSocketShader')
            node_group.interface.new_socket("Color", in_out='INPUT', socket_type='NodeSocketColor')

            input_node = node_group.nodes.new('NodeGroupInput')
            bsdf_node = node_group.nodes.new('ShaderNodeEmission')
            bsdf_node.location.x = 200.0
            output_node = node_group.nodes.new('NodeGroupOutput')
            output_node.location.x = 400.0

            node_group.links.new(bsdf_node.inputs[0], input_node.outputs[0])
            node_group.links.new(output_node.inputs[0], bsdf_node.outputs[0])
        return node_group

    def format(self, tokensource, _outfile):
        """Called by Pygments as part of highlighting."""
        tokensource = tuple(tokensource)

        # Set styles for each character
        current_char = 0
        for ttype, value in tokensource:
            while ttype not in self.styles:
                ttype = ttype.parent
            style = self.styles[ttype]

            for char in value:
                # Can happen for last token
                if current_char >= len(self.obj.data.body_format):
                    return
                self.obj.data.body_format[current_char].use_bold = style["bold"]
                self.obj.data.body_format[current_char].use_italic = style["italic"]
                self.obj.data.body_format[current_char].use_underline = style["underline"]
                self.obj.data.body_format[current_char].material_index = self.get_material_slot(
                    ttype).slot_index
                current_char += 1


class HighlightOperatorMixin(bpy.types.Operator):
    styles = [(style, style.title(), "") for style in pygments.styles.get_all_styles()]
    style: bpy.props.EnumProperty(
        items=styles,
        name="Style",
        description="Color scheme to use for the syntax highlighting",
        default="default",
    )

    languages = []
    for lexer in pygments.lexers.get_all_lexers():
        if lexer[1]:
            alias = lexer[1][0]
        else:
            alias = lexer[0]
        description = lexer[0]
        if lexer[2]:
            description += f', extensions: {", ".join(lexer[2])}'
        languages.append((alias, lexer[0], description))
    del lexer, alias, description
    languages.sort(key=lambda l: l[1].lower())
    language: bpy.props.EnumProperty(
        items=languages,
        name="Language",
        description="Language syntax to use for the highlighting"
    )

    def get_object(self, context):
        raise NotImplementedError

    def execute(self, context):
        mode = None
        if bpy.context.active_object is not None:
            mode = bpy.context.active_object.mode
            bpy.ops.object.mode_set(mode='OBJECT')

        obj = self.get_object(context)
        lexer = pygments.lexers.get_lexer_by_name(self.language, stripall=True)
        pygments.highlight(obj.data.body, lexer, BlenderFormatter(obj, style=self.style))
        obj["highlighter_language"] = self.language
        obj["highlighter_style"] = self.style

        if mode is not None:
            bpy.ops.object.mode_set(mode=mode)
        return {'FINISHED'}


class TEXT_OT_To_Highlighted_Object(HighlightOperatorMixin):
    """Create 3D text object from active text data-block, with syntax highlighting"""
    bl_idname = "text.to_highlighted_object"
    bl_label = "To Syntax-Highlighted 3D Object"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(self, context):
        return (bpy.context.space_data.type == 'TEXT_EDITOR'
                and bpy.context.space_data.text is not None)

    def invoke(self, context, _event):
        text = context.space_data.text
        script_text = "\n".join(line.body for line in text.lines)

        # Try to guess the syntax based on content and filename or datablock name
        if text.filepath:
            filename = os.path.basename(text.filepath)
        else:
            filename = text.name.rstrip(".0123456789")

        try:
            lexer = pygments.lexers.guess_lexer_for_filename(filename, script_text)
        except pygments.util.ClassNotFound:
            lexer = pygments.lexers.guess_lexer(script_text)
        self.language = lexer.aliases[0] if lexer.aliases else lexer.name

        return context.window_manager.invoke_props_dialog(self)

    @staticmethod
    def get_font(filepath):
        if not filepath:
            return None
        # Try getting existing font with this filepath.
        for font in bpy.data.fonts:
            if os.path.abspath(bpy.path.abspath(filepath)) == os.path.abspath(bpy.path.abspath(font.filepath)):
                return font
        # Load it if not found.
        if os.path.isfile(filepath):
            return bpy.data.fonts.load(filepath)

    def get_object(self, context):
        text = context.space_data.text
        name = text.name
        script_text = "\n".join(line.body for line in text.lines)

        text_data = bpy.data.curves.new(name, 'FONT')
        text_data.body = script_text

        obj = bpy.data.objects.new(name, text_data)

        obj.location = context.scene.cursor.location
        context.collection.objects.link(obj)
        context.view_layer.objects.active = obj
        for other_obj in context.scene.objects:
            other_obj.select_set(other_obj == obj)

        addon_prefs = context.preferences.addons[__package__].preferences
        if (font := self.get_font(addon_prefs.regular_font_path)) is not None:
            text_data.font = font
        if (font := self.get_font(addon_prefs.bold_font_path)) is not None:
            text_data.font_bold = font
        if (font := self.get_font(addon_prefs.italic_font_path)) is not None:
            text_data.font_italic = font
        if (font := self.get_font(addon_prefs.bold_italic_font_path)) is not None:
            text_data.font_bold_italic = font

        return obj


class FONT_OT_Highlight_Selected_Text(HighlightOperatorMixin):
    """Apply syntax highlighting on the active object"""
    bl_idname = "font.highlight_active_text"
    bl_label = "Syntax Highlight Object"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(self, context):
        return (bpy.context.active_object is not None
                and bpy.context.active_object.type == 'FONT')

    def invoke(self, context, _event):
        obj = context.active_object
        script_text = obj.data.body

        style = obj.get("highlighter_style")
        if style is not None and style in [s[0] for s in self.styles]:
            self.style = style

        language = obj.get("highlighter_language")
        if language not in [l[0] for l in self.languages]:
            language = None

        if language is None:
            lexer = pygments.lexers.guess_lexer(script_text)
            language = lexer.aliases[0] if lexer.aliases else lexer.name

        self.language = language
        return context.window_manager.invoke_props_dialog(self)

    def get_object(self, context):
        return context.active_object


class HighlighterPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    regular_font_path: bpy.props.StringProperty(
        name="Regular Font Path",
        description="Default path for the regular font",
        subtype='FILE_PATH',
    )
    bold_font_path: bpy.props.StringProperty(
        name="Bold Font Path",
        description="Default path for the bold font",
        subtype='FILE_PATH',
    )
    italic_font_path: bpy.props.StringProperty(
        name="Italic Font Path",
        description="Default path for the italic font",
        subtype='FILE_PATH',
    )
    bold_italic_font_path: bpy.props.StringProperty(
        name="Bold & Italic Font Path",
        description="Default path for the bold and italic font",
        subtype='FILE_PATH',
    )

    def draw(self, context):
        layout = self.layout

        col = layout.column()
        col.label(text="Default Font Paths")
        col.prop(self, "regular_font_path", text="Regular")
        col.prop(self, "bold_font_path", text="Bold")
        col.prop(self, "italic_font_path", text="Italic")
        col.prop(self, "bold_italic_font_path", text="Bold & Italic")


def text_edit_to_3d_menu_func(self, _context):
    self.layout.operator("text.to_highlighted_object", text="Syntax-Highlighted")


def font_object_font_menu_func(self, _context):
    self.layout.separator()
    self.layout.operator("font.highlight_active_text", text="Apply Syntax Highlighting")


def register():
    bpy.utils.register_class(HighlighterPreferences)
    bpy.utils.register_class(TEXT_OT_To_Highlighted_Object)
    bpy.utils.register_class(FONT_OT_Highlight_Selected_Text)
    bpy.types.TEXT_MT_edit_to3d.append(text_edit_to_3d_menu_func)
    bpy.types.VIEW3D_MT_edit_font.append(font_object_font_menu_func)


def unregister():
    bpy.types.VIEW3D_MT_edit_font.remove(font_object_font_menu_func)
    bpy.types.TEXT_MT_edit_to3d.remove(text_edit_to_3d_menu_func)
    bpy.utils.unregister_class(FONT_OT_Highlight_Selected_Text)
    bpy.utils.unregister_class(TEXT_OT_To_Highlighted_Object)
    bpy.utils.unregister_class(HighlighterPreferences)


if __name__ == "__main__":
    register()
